/*******************************************************************************
 *
 * MIT License
 *
 * Copyright (c) 2024 Advanced Micro Devices, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *******************************************************************************/

#include "rocrand_wrapper.hpp"

#include <rocrand/rocrand.hpp>

namespace gpumemrand {

namespace {
template <typename T>
int gen_0_1_impl(T* buf, size_t sz)
{
    rocrand_cpp::xorwow_engine<> g;
    rocrand_cpp::uniform_real_distribution<T> d;
    d(g, buf, sz);
    return 0;
}
} // namespace

int gen_0_1(double* buf, size_t sz) //
{
    return gen_0_1_impl(buf, sz);
}

int gen_0_1(float* buf, size_t sz) //
{
    return gen_0_1_impl(buf, sz);
}

int gen_0_1(half_float::half* buf, size_t sz) //
{
    return gen_0_1_impl(reinterpret_cast<half*>(buf), sz);
}

} // namespace gpumemrand
