/*******************************************************************************
 *
 * MIT License
 *
 * Copyright (c) 2017 Advanced Micro Devices, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *******************************************************************************/
#ifndef GUARD_MIOPEN_CONVOLUTION_FFT_HPP_
#define GUARD_MIOPEN_CONVOLUTION_FFT_HPP_

namespace miopen {

struct FFTConvParams
{
    static int TileSize(int in_h, int in_w)
    {
        int NX, NY; // fft tile width, height

        if((in_h == 7) && (in_w == 7))
        {
            NY = 12;
            NX = 12;
        }
        else if((in_h == 14) && (in_w == 14))
        {
            NY = 18;
            NX = 18;
        }
        else
        {
            NY = 32;
            NX = 32;
        }

        return NY * (1 + NX / 2);
    }

    static const int TransposePadding = 64;
    static const int NumKernels       = 7;
};

} // namespace miopen

#endif // GUARD_MIOPEN_CONVOLUTION_FFT_HPP_
