/*******************************************************************************
 *
 * MIT License
 *
 * Copyright (c) 2019 Advanced Micro Devices, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *******************************************************************************/

#pragma once

#include <string>

namespace miopen {

struct NetworkConfig
{
    NetworkConfig() = default;
    explicit NetworkConfig(const std::string& value_) : value(value_) {}
    explicit NetworkConfig(std::string&& value_) noexcept : value(std::move(value_)) {}
    operator std::string() const { return value; }
    const std::string& ToString() const { return value; }

private:
    std::string value;
};

struct AlgorithmName
{
    AlgorithmName() = default;
    explicit AlgorithmName(const std::string& value_) : value(value_) {}
    operator std::string() const { return value; }
    const std::string& ToString() const { return value; }

    bool operator<(const AlgorithmName& r) const { return (value < r.value); }

private:
    std::string value;
};

} // namespace miopen
